@@ -25,7 +25,7 @@
 -export([record/3,function/3]).
 
 -export([comp_define/1]).
--export([lift_func/2,lift_expr/3,ivars_expr/1]).
+-export([lift_func/2,lift_expr/3,lift_expr/4,ivars_expr/1]).
 
 -export([test/1]).
 
@@ -42,6 +42,10 @@
              fc=0                               %Local function index
             }).
 
+-record(lift, {type,                            %Lift type
+               name,                            %New lifted name
+               ivars}).                         %Imported variables
+
 %% comp_define(DefForm) -> Funcs
 
 comp_define({Name,Def,Line}) ->
@@ -54,7 +58,8 @@ comp_define({Name,Def,Line}) ->
 
 record(Name, Fs, Line) ->
     St0 = #cl{func=Name,arity=record,line=Line,vc=0,fc=0},
-    {Lfs,Fncs,St1} = lift_rec_fields(Fs, [], St0),
+    Lifts = orddict:new(),
+    {Lfs,Fncs,St1} = lift_rec_fields(Fs, Lifts, [], St0),
     {Lfncs,[],_} = lift_loop(Fncs, St1),
     {Lfs,Lfncs}.
 
@@ -89,7 +94,8 @@ lift_loop(Funcs0, St0) ->
 %%  the resulting functions.
 
 lift_func({Name,Def0,L}, St0) ->
-    {Def1,Lds,St1} = lift_expr(Def0, [], St0),
+    Lifts = orddict:new(),
+    {Def1,Lds,St1} = lift_expr(Def0, Lifts, [], St0),
     {{Name,Def1,L},Lds,St1}.
 
 lift_funcs(Defs, St) ->
@@ -100,196 +106,254 @@ lift_funcs(Defs, St) ->
     lists:foldl(Fun, {[],[],St}, Defs).
 
 %% lift_expr(Expr, LocalDefs, State) -> {AST,LocalDefs,State}.
-%%  Lambda lift the local functions in an expression.
+%% lift_expr(Expr, LiftedFuncs, LocalDefs, State) -> {AST,LocalDefs,State}.
+%%  Lambda lift the local functions in an expression. The Lifts are
+%%  the current function lifts which need to be done.
+
+lift_expr(Expr, Lds, St) ->
+    lift_expr(Expr, [], Lds, St).
 
 %% Core data special forms.
-lift_expr(?Q(E), Lds, St) -> {?Q(E),Lds,St};
+lift_expr(?Q(E), _Lifts, Lds, St) -> {?Q(E),Lds,St};
 %% Record special forms.
-lift_expr(['record',Name|Args], Lds0, St0) ->
-    {Largs,Lds1,St1} = lift_rec_args(Args, Lds0, St0),
+lift_expr(['record',Name|Args], Lifts, Lds0, St0) ->
+    {Largs,Lds1,St1} = lift_rec_args(Args, Lifts, Lds0, St0),
     {['record',Name|Largs],Lds1,St1};
 %% make-record has been deprecated but we sill accept it for now.
-lift_expr(['make-record',Name|Args], Lds0, St0) ->
-    {Largs,Lds1,St1} = lift_rec_args(Args, Lds0, St0),
+lift_expr(['make-record',Name|Args], Lifts, Lds0, St0) ->
+    {Largs,Lds1,St1} = lift_rec_args(Args, Lifts, Lds0, St0),
     {['make-record',Name|Largs],Lds1,St1};
-lift_expr(['is-record',E,Name], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
+lift_expr(['is-record',E,Name], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
     {['is-record',Le,Name],Lds1,St1};
-lift_expr(['record-index',_Name,_F]=Ri, Lds, St) ->
+lift_expr(['record-index',_Name,_F]=Ri, _Lifts, Lds, St) ->
     {Ri,Lds,St};
-lift_expr(['record-field',E,Name,F], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
+lift_expr(['record-field',E,Name,F], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
     {['record-field',Le,Name,F],Lds1,St1};
-lift_expr(['record-update',E,Name|Args], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
-    {Largs,Lds2,St2} = lift_rec_args(Args, Lds1, St1),
+lift_expr(['record-update',E,Name|Args], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
+    {Largs,Lds2,St2} = lift_rec_args(Args, Lifts, Lds1, St1),
     {['record-update',Le,Name|Largs],Lds2,St2};
 %% Struct special forms.
-lift_expr(['struct',Name|Args], Lds0, St0) ->
-    {Largs,Lds1,St1} = lift_rec_args(Args, Lds0, St0),
+lift_expr(['struct',Name|Args], Lifts, Lds0, St0) ->
+    {Largs,Lds1,St1} = lift_rec_args(Args, Lifts, Lds0, St0),
     {['struct',Name|Largs],Lds1,St1};
-lift_expr(['is-struct',E], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
+lift_expr(['is-struct',E], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
     {['is-struct',Le],Lds1,St1};
-lift_expr(['is-struct',E,Name], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
+lift_expr(['is-struct',E,Name], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
     {['is-struct',Le,Name],Lds1,St1};
-lift_expr(['struct-field',E, Name,F], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
+lift_expr(['struct-field',E, Name,F], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
     {['struct-field',Le,Name,F],Lds1,St1};
-lift_expr(['struct-update',E,Name|Args], Lds0, St0) ->
-    {Le,Lds1,St1} = lift_expr(E, Lds0, St0),
-    {Largs,Lds2,St2} = lift_rec_args(Args, Lds1, St1),
+lift_expr(['struct-update',E,Name|Args], Lifts, Lds0, St0) ->
+    {Le,Lds1,St1} = lift_expr(E, Lifts, Lds0, St0),
+    {Largs,Lds2,St2} = lift_rec_args(Args, Lifts, Lds1, St1),
     {['struct-update',Le,Name|Largs],Lds2,St2};
 %% Function forms.
-lift_expr([function,_,_]=Func, Lds, St) ->
-    {Func,Lds,St};
-lift_expr([function,_,_,_]=Func, Lds, St) ->
+lift_expr([function,Name,Arity], Lifts, Lds, St) ->
+    lift_function_ref(Name, Arity, Lifts, Lds, St);
+lift_expr([function,_,_,_]=Func, _Lifts, Lds, St) ->
     {Func,Lds,St};
 %% Core closure special forms.
-lift_expr([lambda,Args|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+lift_expr([lambda,Args|Body0], Lifts, Lds0, St0) ->
+    {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
     {[lambda,Args|Body1],Lds1,St1};
-lift_expr(['match-lambda'|Cls0], Lds0, St0) ->
-    {Cls1,Lds1,St1} = lift_cls(Cls0, Lds0, St0),
+lift_expr(['match-lambda'|Cls0], Lifts, Lds0, St0) ->
+    {Cls1,Lds1,St1} = lift_cls(Cls0, Lifts, Lds0, St0),
     {['match-lambda'|Cls1],Lds1,St1};
-lift_expr(['let',Vbs|Body], Lds, St) ->
-    lift_let(Vbs, Body, Lds, St);
-lift_expr(['let-function',Fbs|Body], Lds, St) ->
-    lift_let_function(Fbs, Body, Lds, St);
-lift_expr(['letrec-function',Fbs|Body], Lds, St) ->
-    lift_letrec_function(Fbs, Body, Lds, St);
+lift_expr(['let',Vbs|Body], Lifts, Lds, St) ->
+    lift_let(Vbs, Body, Lifts, Lds, St);
+lift_expr(['let-function',Fbs|Body], Lifts, Lds, St) ->
+    lift_let_function(Fbs, Body, Lifts, Lds, St);
+lift_expr(['letrec-function',Fbs|Body], Lifts, Lds, St) ->
+    lift_letrec_function(Fbs, Body, Lifts, Lds, St);
 %% Core control special forms.
-lift_expr([progn|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+lift_expr([progn|Body0], Lifts, Lds0, St0) ->
+    {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
     {[progn|Body1],Lds1,St1};
-lift_expr(['if'|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+lift_expr(['if'|Body0], Lifts, Lds0, St0) ->
+    {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
     {['if'|Body1],Lds1,St1};
-lift_expr(['case',Expr|Cls], Lds, St) ->
-    lift_case(Expr, Cls, Lds, St);
-lift_expr(['catch'|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+lift_expr(['case',Expr|Cls], Lifts, Lds, St) ->
+    lift_case(Expr, Cls, Lifts, Lds, St);
+lift_expr(['maybe'|Body], Lifts, Lds, St) ->
+    lift_maybe(Body, Lifts, Lds, St);
+lift_expr(['catch'|Body0], Lifts, Lds0, St0) ->
+    {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
     {['catch'|Body1],Lds1,St1};
-lift_expr(['try'|Try], Lds, St) ->
-    lift_try(Try, Lds, St);
-lift_expr([funcall|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
-    {[funcall|Body1],Lds1,St1};
+lift_expr(['try'|Try], Lifts, Lds, St) ->
+    lift_try(Try, Lifts, Lds, St);
+lift_expr([funcall|Args0], Lifts, Lds0, St0) ->
+    {Args1,Lds1,St1} = lift_exprs(Args0, Lifts, Lds0, St0),
+    {[funcall|Args1],Lds1,St1};
 %% List/binary comprehensions.
-lift_expr(['lc',Qs,E], Lds, St) ->
-    lift_comp('lc', Qs, E, Lds, St);
-lift_expr(['list-comp',Qs,E], Lds, St) ->
-    lift_comp('list-comp', Qs, E, Lds, St);
-lift_expr(['bc',Qs,E], Lds, St) ->
-    lift_comp('bc', Qs, E, Lds, St);
-lift_expr(['binary-comp',Qs,E], Lds, St) ->
-    lift_comp('binary-comp', Qs, E, Lds, St);
+lift_expr(['lc',Qs,E], Lifts, Lds, St) ->
+    lift_comp('lc', Qs, E, Lifts, Lds, St);
+lift_expr(['list-comp',Qs,E], Lifts, Lds, St) ->
+    lift_comp('list-comp', Qs, E, Lifts, Lds, St);
+lift_expr(['bc',Qs,E], Lifts, Lds, St) ->
+    lift_comp('bc', Qs, E, Lifts, Lds, St);
+lift_expr(['binary-comp',Qs,E], Lifts, Lds, St) ->
+    lift_comp('binary-comp', Qs, E, Lifts, Lds, St);
 %% Finally the general cases.
-lift_expr([call|Body0], Lds0, St0) ->
-    {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
-    {[call|Body1],Lds1,St1};
-lift_expr([Func|Args0], Lds0, St0) when is_atom(Func) ->
-    {Args1,Lds1,St1} = lift_exprs(Args0, Lds0, St0),
-    {[Func|Args1],Lds1,St1};
-lift_expr(Lit, Lds, St) -> {Lit,Lds,St}.
+lift_expr([call|Args0], Lifts, Lds0, St0) ->
+    {Args1,Lds1,St1} = lift_exprs(Args0, Lifts, Lds0, St0),
+    {[call|Args1],Lds1,St1};
+lift_expr([Func|Args], Lifts, Lds, St) when is_atom(Func) ->
+    lift_func_call(Func, Args, Lifts, Lds, St);
+%% Everything else is a literal.
+lift_expr(Lit, _Lifts, Lds, St) -> {Lit,Lds,St}.
 
-lift_exprs(Exprs, Lds, St) ->
+lift_func_call(Name, Args0, Lifts, Lds0, St0) ->
+    %% Most of the core data special forms can be handled here as well.
+    {Args1,Lds1,St1} = lift_exprs(Args0, Lifts, Lds0, St0),
+    Arity = length(Args1),
+    Call = case lifted_function(Name, Arity, Lifts) of
+               {yes,#lift{type=call,name=NewName,ivars=Ivars}} ->
+                   [NewName | Args1 ++ Ivars];
+               {yes,#lift{type=apply,name=NewName,ivars=Ivars}} ->
+                   [funcall,NewName | Args1 ++ Ivars];
+               no ->
+                   [Name | Args1]
+           end,
+    {Call,Lds1,St1}.
+
+lift_exprs(Exprs, Lifts, Lds, St) ->
     Fun = fun (E0, {Es,Lds0,St0}) ->
-                  {E1,Lds1,St1} = lift_expr(E0, Lds0, St0),
+                  {E1,Lds1,St1} = lift_expr(E0, Lifts, Lds0, St0),
                   {[E1|Es],Lds1,St1}
           end,
     lists:foldr(Fun, {[],Lds,St}, Exprs).
 
-lift_rec_fields([[F,V|Type]|Fs], Lds0, St0) ->
-    {Lv,Lds1,St1} = lift_expr(V, Lds0, St0),
-    {Lfs,Lds2,St2} = lift_rec_fields(Fs, Lds1, St1),
+lift_rec_fields([[F,V|Type]|Fs], Lifts, Lds0, St0) ->
+    {Lv,Lds1,St1} = lift_expr(V, Lifts, Lds0, St0),
+    {Lfs,Lds2,St2} = lift_rec_fields(Fs, Lifts, Lds1, St1),
     {[[F,Lv|Type]|Lfs],Lds2,St2};
-lift_rec_fields([F|Fs], Lds0, St0) ->
-    {Lfs,Lds1,St1} = lift_rec_fields(Fs, Lds0, St0),
+lift_rec_fields([F|Fs], Lifts, Lds0, St0) ->
+    {Lfs,Lds1,St1} = lift_rec_fields(Fs, Lifts, Lds0, St0),
     {[F|Lfs],Lds1,St1};
-lift_rec_fields([], Lds, St) -> {[],Lds,St}.
+lift_rec_fields([], _Lifts, Lds, St) -> {[],Lds,St}.
 
-lift_rec_args([F,V|As], Lds0, St0) ->
-    {Lv,Lds1,St1} = lift_expr(V, Lds0, St0),
-    {Las,Lds2,St2} = lift_rec_args(As, Lds1, St1),
+lift_rec_args([F,V|As], Lifts, Lds0, St0) ->
+    {Lv,Lds1,St1} = lift_expr(V, Lifts, Lds0, St0),
+    {Las,Lds2,St2} = lift_rec_args(As, Lifts, Lds1, St1),
     {[F,Lv|Las],Lds2,St2};
-lift_rec_args([], Lds, St) -> {[],Lds,St}.
+lift_rec_args([], _Lifts, Lds, St) -> {[],Lds,St}.
 
-lift_let(Vbs0, Body0, Lds0, St0) ->
+%% lift_function_ref(Name, Arity, Lifts, LocalDefs, State) ->
+%%     {Lifted,LocalDefs,State}.
+%%  Check if [function,Name,Arity] needs to tbe lifted and whether
+%%  Ivars force it to become a lambda.
+
+lift_function_ref(Name, Arity, Lifts, Lds, St) ->
+    Lfunc = case lifted_function(Name, Arity, Lifts) of
+                {yes,#lift{name=NewName,ivars=Ivars}} ->
+                    if length(Ivars) > 0 ->
+                            Vars = new_vars(Arity),
+                            [lambda,Vars,[funcall,?Q(NewName) | Vars ++ Ivars]];
+                       true ->
+                            [function,NewName,Arity]
+                    end;
+                no ->                   %When not lifted
+                    [function,Name,Arity]
+            end,
+    {Lfunc,Lds,St}.
+
+%% lift_let(VarBindings, Body, LiftedFuncs, LocalDefines, State) ->
+%%     {Let,LocalDefines,State}.
+
+lift_let(Vbs0, Body0, Lifts, Lds0, St0) ->
+    {Vbs1,Lds1,St1} = lift_let_bindings(Vbs0, Lifts, Lds0, St0),
+    {Body1,Lds2,St2} = lift_exprs(Body0, Lifts, Lds1, St1),
+    {['let',Vbs1|Body1],Lds2,St2}.
+
+lift_let_bindings(Vbs0, Lifts, Lds0, St0) ->
     Fun = fun ([Pat,['when'|_]=G,Expr0], {Ldsa,Sta}) ->
-                  {Expr1,Ldsb,Stb} = lift_expr(Expr0, Ldsa, Sta),
+                  {Expr1,Ldsb,Stb} = lift_expr(Expr0, Lifts, Ldsa, Sta),
                   {[Pat,G,Expr1],{Ldsb,Stb}};
               ([Pat,Expr0], {Ldsa,Sta}) ->
-                  {Expr1,Ldsb,Stb} = lift_expr(Expr0, Ldsa, Sta),
+                  {Expr1,Ldsb,Stb} = lift_expr(Expr0, Lifts, Ldsa, Sta),
                   {[Pat,Expr1],{Ldsb,Stb}}
           end,
     {Vbs1,{Lds1,St1}} = lists:mapfoldl(Fun, {Lds0,St0}, Vbs0),
-    {Body1,Lds2,St2} = lift_exprs(Body0, Lds1, St1),
-    {['let',Vbs1|Body1],Lds2,St2}.
+    {Vbs1,Lds1,St1}.
 
-%% lift_let_function(FuncBindings, LocalDefines, State) ->
+%% lift_let_function(FuncBindings, Body, LiftedFuncs, LocalDefines, State) ->
 %%     {LocalBody,LocalDefines,State}.
 %%  We can check imported vars separately for each local function as
 %%  they do not know of each other.
 
-lift_let_function(Fbs0, Body0, Lds0, St0) ->
+lift_let_function(Fbs0, Body, Lifts0, Lds0, St0) ->
     %% Build new name function binding and name transform data.
     Line = St0#cl.line,
-    Nfun = fun ([Name,Def0], Ts, Sta) ->
+    Nfun = fun ([Name,Def0], {Ls0,Lda,Sta}) ->
                    Ar = func_arity(Def0),
-                   {New,Stb} = new_local_fun_name(Name, Ar, Sta),
-                   %% Get the imported variables.
+                   %% Get imported variables and append them in definition.
                    Ivs  = ivars_expr(Def0, [], []),
                    Def1 = append_ivars(Def0, Ivs),
-                   {{New,Def1,Line},[{trans,Name,Ar,New,Ivs}|Ts],Stb}
+                   {Def2,Ldb,Stb} = lift_expr(Def1, Lifts0, Lda, Sta),
+                   {New,Ls1,Stc} = lift_function(Name, Ar, call, Ivs, Ls0, Stb),
+                   {{New,Def2,Line},{Ls1,Ldb,Stc}}
            end,
     %% Transform calls in the body.
-    {Fbs1,Trans,St1} = mapfoldl2(Nfun, [], St0, Fbs0),
-    Lds1 = Fbs1 ++ Lds0,
-    %% Apply tranformations to Body.
-    Bfun = fun ({trans,Name,Ar,New,Ivs}, B) ->
-                   trans_expr(B, Name, Ar, New, Ivs)
-           end,
-    Body1 = lists:foldl(Bfun, [progn|Body0], Trans),
-    lift_expr(Body1, Lds1, St1).
+    {Fbs1,{Lifts1,Lds1,St1}} = lists:mapfoldl(Nfun, {Lifts0,Lds0,St0}, Fbs0),
+    %% io:format("lf ~p\n   ~p\b", [Fbs1,Lds1]),
+    Lds2 = Fbs1 ++ Lds1,
+    lift_expr([progn | Body], Lifts1, Lds2, St1).
 
-%% lift_letrec_function(FuncBindings, LocalDefines, State) ->
+%% lift_letrec_function(FuncBindings, Body, LiftedFuncs, LocalDefines, State) ->
 %%     {LocalBody,LocalDefines,State}.
 %%  We cheat a bit when checking imported vars, we just take the union
 %%  of the variables from all the local functions and pass them to all
 %%  functions.
 
-lift_letrec_function(Fbs0, Body0, Lds0, St0) ->
-    %% Get the imported variables.
+lift_letrec_function(Fbs0, Body, Lifts0, Lds0, St0) ->
+    %% Get all the imported variables.
     Ifun = fun ([_,Def], Ivs) -> ivars_expr(Def, [], Ivs) end,
     Ivars = lists:foldl(Ifun, [], Fbs0),
-    %% Build new name function binding and name transform data.
+    %% io:format("lrf ~p\n", [Ivars]),
+    %% Lift the local defined functions and get their new names and data.
+    Lfun = fun ([Name,Def], {Lsa,Sta}) ->
+                   Ar = func_arity(Def),
+                   {New,Lsb,Stb} = lift_function(Name, Ar, call, Ivars, Lsa, Sta),
+                   {{New,Def},{Lsb,Stb}}
+           end,
+    {Fbs1,{Lifts1,St1}} = lists:mapfoldl(Lfun, {Lifts0,St0}, Fbs0),
+    %% Lift the local functions with the joint lifted local functions.
     Line = St0#cl.line,
-    Nfun = fun ([Name,Def0], Ts, Sta) ->
-                   Ar = func_arity(Def0),
-                   {New,Stb} = new_local_fun_name(Name, Ar, Sta),
+    Efun = fun ({New,Def0}, {Lda,Sta}) ->
                    Def1 = append_ivars(Def0, Ivars),
-                   {{New,Def1,Line},[{trans,Name,Ar,New}|Ts],Stb}
+                   {Def2,Ldb,Stb} = lift_expr(Def1, Lifts1, Lda, Sta),
+                   {{New,Def2,Line}, {Ldb,Stb}}
            end,
-    {Fbs1,Trans,St1} = mapfoldl2(Nfun, [], St0, Fbs0),
-    %% Transform calls in the letrec form.
-    Tfun = fun ({trans,Name,Ar,New}, Fbs) ->
-                   Ffun = fun ({Nn,Def0,L}) ->
-                                  Def1 = trans_expr(Def0, Name, Ar, New, Ivars),
-                                  {Nn,Def1,L}
-                          end,
-                   lists:map(Ffun, Fbs)
-           end,
-    Fbs2 = lists:foldl(Tfun, Fbs1, Trans),
-    Lds1 = Fbs2 ++ Lds0,
-    %% Apply tranformations to Body.
-    Bfun = fun ({trans,Name,Ar,New}, B) ->
-                   trans_expr(B, Name, Ar, New, Ivars)
-           end,
-    Body1 = lists:foldl(Bfun, [progn|Body0], Trans),
-    {Body2,Lds2,St2} = lift_expr(Body1, Lds1, St1),
-    {Body2,Lds2,St2}.
+    {Fbs2,{Lds1,St2}} = lists:mapfoldl(Efun, {Lds0,St1}, Fbs1),
+    Lds2 = Fbs2 ++ Lds1,
+    lift_expr([progn | Body], Lifts1, Lds2, St2).
+
+%% lift_function(Name, Arity, LiftType, Ivars, Lifts, State) ->
+%%     {NewName,Lifts,State}.
+%% lifted_function(Name, Arity, Lifts) -> {yes,Lift} | no.
+%%  Lift a function, and check whether a function has been liftes and
+%%  return irs lift data.
+
+lift_function(Name, Arity, Type, Ivars, Lifts0, St0) ->
+    {NewName,St1} = new_local_fun_name(Name, Arity, St0),
+    Lift = #lift{type=Type,name=NewName,ivars=Ivars},
+    Lifts1 = orddict:store({Name,Arity}, Lift, Lifts0),
+    {NewName,Lifts1,St1}.
+    
+lifted_function(Name, Arity, Lifts) ->
+    case orddict:find({Name,Arity}, Lifts) of
+        {ok,Value} -> {yes,Value};
+        error -> no
+    end.
+                       
+func_arity([lambda,Args|_]) -> length(Args);
+func_arity(['match-lambda',[Pats|_]|_]) ->
+    length(Pats).
 
 append_ivars([lambda,Args|Body], Ivars) ->
     [lambda,Args ++ Ivars|Body];
@@ -298,298 +362,108 @@ append_ivars(['match-lambda'|Cls0], Ivars) ->
     Cls1 = lists:map(Fun, Cls0),
     ['match-lambda'|Cls1].
 
-lift_cls(Cls, Lds, St) ->
+lift_cls(Cls, Lifts, Lds, St) ->
     Fun = fun ([Pats,['when'|_]=G|Body0], {Cls0,Lds0,St0}) ->
-                  {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+                  {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
                   {[[Pats,G|Body1]|Cls0],Lds1,St1};
               ([Pats|Body0], {Cls0,Lds0,St0}) ->
-                  {Body1,Lds1,St1} = lift_exprs(Body0, Lds0, St0),
+                  {Body1,Lds1,St1} = lift_exprs(Body0, Lifts, Lds0, St0),
                   {[[Pats|Body1]|Cls0],Lds1,St1}
           end,
     lists:foldr(Fun, {[],Lds,St}, Cls).         %From the right!
 
-lift_case(Expr0, Cls0, Lds0, St0) ->
-    {Expr1,Lds1,St1} = lift_expr(Expr0, Lds0, St0),
-    {Cls1,Lds2,St2} = lift_cls(Cls0, Lds1, St1),
+lift_case(Expr0, Cls0, Lifts, Lds0, St0) ->
+    {Expr1,Lds1,St1} = lift_expr(Expr0, Lifts, Lds0, St0),
+    {Cls1,Lds2,St2} = lift_cls(Cls0, Lifts, Lds1, St1),
     {['case',Expr1|Cls1],Lds2,St2}.
 
-%% lift_try(TryBody, LocalDefs, State) -> {TryBody,LocalDefs,State}.
+%% lift_maybe(Body, LifteFuncs, LocalDefines, State) ->
+%%     {Maybe,LocalDefines,State}.
+%%  We must also explicitly handle let and explicitly lift their
+%%  bodies as maybe bodies.
+
+lift_maybe(Body0, Lifts, Lds0, St0) ->
+    {Body1,Lds1,St1} = lift_maybe_body(Body0, Lifts, Lds0, St0),
+    {['maybe'|Body1],Lds1,St1}.
+
+lift_maybe_body([['?=',Pat,Expr0]|Mes0], Lifts, Lds0, St0) ->
+    {Expr1,Lds1,St1} = lift_expr(Expr0, Lifts, Lds0, St0),
+    {Mes1,Lds2,St2} = lift_maybe_body(Mes0, Lifts, Lds1, St1),
+    {[['?=',Pat,Expr1]|Mes1],Lds2,St2};
+lift_maybe_body([['let',Vbs|Body]|Mes0], Lifts, Lds0, St0) ->
+    {Let,Lds1,St1} = lift_maybe_let(Vbs, Body, Lifts, Lds0, St0),
+    {Mes1,Lds2,St2} = lift_maybe_body(Mes0, Lifts, Lds1, St1),
+    {[Let|Mes1],Lds2,St2};
+lift_maybe_body([['else'|Cls0]], Lifts, Lds0, St0) ->
+    {Cls1,Lds1,St1} = lift_cls(Cls0, Lifts, Lds0, St0),
+    {[['else'|Cls1]],Lds1,St1};
+lift_maybe_body([Expr0|Mes0], Lifts, Lds0, St0) ->
+    {Expr1,Lds1,St1} = lift_expr(Expr0, Lifts, Lds0, St0),
+    {Mes1,Lds2,St2} = lift_maybe_body(Mes0, Lifts, Lds1, St1),
+    {[Expr1|Mes1],Lds2,St2};
+lift_maybe_body([], _Lift, Lds, St) ->
+    {[],Lds,St}.
+
+lift_maybe_let(Vbs0, Body0, Lifts, Lds0, St0) ->
+    {Vbs1,Lds1,St1} = lift_let_bindings(Vbs0, Lifts, Lds0, St0),
+    {Body1,Lds2,St2} = lift_maybe_body(Body0, Lifts, Lds1, St1),
+    {['let',Vbs1|Body1],Lds2,St2}.
+
+%% lift_try(TryBody, LiftedFuncs, LocalDefs, State) ->
+%%     {TryBody,LocalDefs,State}.
 %%  Step down the try body lifting the local functions.
 
-lift_try(Try0, Lds0, St0) ->
+lift_try(Try0, Lifts, Lds0, St0) ->
     Fun = fun (T0, {L0,S0}) ->
-                  {T1,L1,S1} = lift_try_1(T0, L0, S0),
+                  {T1,L1,S1} = lift_try_1(T0, Lifts, L0, S0),
                   {T1,{L1,S1}}
           end,
     {Try1,{Lds1,St1}} = lists:mapfoldl(Fun, {Lds0,St0}, Try0),
     {['try'|Try1],Lds1,St1}.
 
-lift_try_1(['case'|Case0], Lds0, St0) ->
-    {Case1,Lds1,St1} = lift_cls(Case0, Lds0, St0),
+lift_try_1(['case'|Case0], Lifts, Lds0, St0) ->
+    {Case1,Lds1,St1} = lift_cls(Case0, Lifts, Lds0, St0),
     {['case'|Case1],Lds1,St1};
-lift_try_1(['catch'|Catch0], Lds0, St0) ->
-    {Catch1,Lds1,St1} = lift_cls(Catch0, Lds0, St0),
+lift_try_1(['catch'|Catch0], Lifts, Lds0, St0) ->
+    {Catch1,Lds1,St1} = lift_cls(Catch0, Lifts, Lds0, St0),
     {['catch'|Catch1],Lds1,St1};
-lift_try_1(['after'|After0], Lds0, St0) ->
-    {After1,Lds1,St1} = lift_exprs(After0, Lds0, St0),
+lift_try_1(['after'|After0], Lifts, Lds0, St0) ->
+    {After1,Lds1,St1} = lift_exprs(After0, Lifts, Lds0, St0),
     {['after'|After1],Lds1,St1};
-lift_try_1(E, Lds, St) ->                       %The try expression.
-    lift_expr(E, Lds, St).
+lift_try_1(E, Lifts, Lds, St) ->                %The try expression.
+    lift_expr(E, Lifts, Lds, St).
 
-%% lift_comp(Commprehension, Qualifiers, Expr, LocalDefs, State) ->
-%%   {Comprehension,LocalDefs,State}.
+%% lift_comp(Commprehension, Qualifiers, Expr, LiftedFuncs, LocalDefs, State) ->
+%%     {Comprehension,LocalDefs,State}.
 %%  Lift comprehensions. Only the expressions in the comprehensions
 %%  need to be lifted, no guards or patterns.
 
-lift_comp(Comp, Qs0, E0, Lds0, St0) ->
+lift_comp(Comp, Qs0, E0, Lifts, Lds0, St0) ->
     %% io:format("lc ~p\n", [[Comp,Qs0,E0]]),
-    {Qs1,Lds1,St1} = lift_comp_quals(Qs0, Lds0, St0),
-    {E1,Lds2,St2} = lift_expr(E0, Lds1, St1),
+    {Qs1,Lds1,St1} = lift_comp_quals(Qs0, Lifts, Lds0, St0),
+    {E1,Lds2,St2} = lift_expr(E0, Lifts, Lds1, St1),
     {[Comp,Qs1,E1],Lds2,St2}.
 
-lift_comp_quals(Qs, Lds, St) ->
+lift_comp_quals(Qs, Lifts, Lds, St) ->
     lists:foldr(fun (Q0, {Qs0,Lds0,St0}) ->
-			{Q1,Lds1,St1} = lift_comp_qual(Q0, Lds0, St0),
-			{[Q1|Qs0],Lds1,St1}
-		end, {[],Lds,St}, Qs).
+                        {Q1,Lds1,St1} = lift_comp_qual(Q0, Lifts, Lds0, St0),
+                        {[Q1|Qs0],Lds1,St1}
+                end, {[],Lds,St}, Qs).
 
-lift_comp_qual(['<-',Pat,E0], Lds0, St0) ->
-    {E1,Lds1,St1} = lift_expr(E0, Lds0, St0),
+lift_comp_qual(['<-',Pat,E0], Lifts, Lds0, St0) ->
+    {E1,Lds1,St1} = lift_expr(E0, Lifts, Lds0, St0),
     {['<-',Pat,E1],Lds1,St1};
-lift_comp_qual(['<-',Pat,G,E0], Lds0, St0) ->
-    {E1,Lds1,St1} = lift_expr(E0, Lds0, St0),
+lift_comp_qual(['<-',Pat,G,E0], Lifts, Lds0, St0) ->
+    {E1,Lds1,St1} = lift_expr(E0, Lifts, Lds0, St0),
     {['<-',Pat,G,E1],Lds1,St1};
-lift_comp_qual(['<=',Pat,E0], Lds0, St0) ->
-    {E1,Lds1,St1} = lift_expr(E0, Lds0, St0),
+lift_comp_qual(['<=',Pat,E0], Lifts, Lds0, St0) ->
+    {E1,Lds1,St1} = lift_expr(E0, Lifts, Lds0, St0),
     {['<=',Pat,E1],Lds1,St1};
-lift_comp_qual(['<=',Pat,G,E0], Lds0, St0) ->
-    {E1,Lds1,St1} = lift_expr(E0, Lds0, St0),
+lift_comp_qual(['<=',Pat,G,E0], Lifts, Lds0, St0) ->
+    {E1,Lds1,St1} = lift_expr(E0, Lifts, Lds0, St0),
     {['<=',Pat,G,E1],Lds1,St1};
-lift_comp_qual(Test, Lds, St) ->
-    lift_expr(Test, Lds, St).
-
-%% trans_expr(Call, OldName, Arity, NewName, ImportedVars) -> Expr.
-%%  Translate function call from old Name to New and add imported
-%%  variables.
-
-%% Core data special forms.
-trans_expr(?Q(E), _, _, _, _) -> ?Q(E);
-trans_expr([binary|Segs0],  Old, Ar, New, Ivars) ->
-    Segs1 = trans_bitsegs(Segs0, Old, Ar, New, Ivars),
-    [binary|Segs1];
-%% Record forms.
-trans_expr(['record',Rname|Args], Old, Ar, New, Ivars) ->
-    Targs = trans_rec_args(Args, Old, Ar, New, Ivars),
-    ['record',Rname|Targs];
-%% make-record has been deprecated but we sill accept it for now.
-trans_expr(['make-record',Rname|Args], Old, Ar, New, Ivars) ->
-    Targs = trans_rec_args(Args, Old, Ar, New, Ivars),
-    ['make-record',Rname|Targs];
-trans_expr(['is-record',E,Rname], Old, Ar, New, Ivars) ->
-    Te = trans_expr(E, Old, Ar, New, Ivars),
-    ['is-record',Te,Rname];
-trans_expr(['record-index',_Name,_F]=Ri, _, _, _, _) ->
-    Ri;                                         %Nothing to do here
-trans_expr(['record-field',E,Rname,F], Old, Ar, New, Ivars) ->
-    Te = trans_expr(E, Old, Ar, New, Ivars),
-    ['record-field',Te,Rname,F];
-trans_expr(['record-update',E,Rname|Args], Old, Ar, New, Ivars) ->
-    Te = trans_expr(E, Old, Ar, New, Ivars),
-    Targs = trans_rec_args(Args, Old, Ar, New, Ivars),
-    ['record-update',Te,Rname|Targs];
-%% Function forms.
-trans_expr([function,F,A]=Func, Old, Ar, New, Ivars) ->
-    if F =:= Old, A =:= Ar ->
-            %% Must return a function of arity A here which calls the
-            %% lifted functions! Can access the imported variables.
-            Vars = new_vars(A),
-            [lambda,Vars,[New|Vars++Ivars]];
-       true ->
-            Func
-    end;
-trans_expr([function,_,_,_]=Func, _, _, _, _) ->
-    Func;                                       %Nothing to do here
-%% Core closure special forms.
-trans_expr([lambda,Args|Body0], Name, Ar, New, Ivars) ->
-    Body1 = trans_exprs(Body0, Name, Ar, New, Ivars),
-    [lambda,Args|Body1];
-trans_expr(['match-lambda'|Cls0], Name, Ar, New, Ivars) ->
-    Cls1 = trans_cls(Cls0, Name, Ar, New, Ivars),
-    ['match-lambda'|Cls1];
-trans_expr(['let',Vbs|Body], Name, Ar, New, Ivars) ->
-    trans_let(Vbs, Body, Name, Ar, New, Ivars);
-trans_expr(['let-function',Fbs|Body], Name, Ar, New, Ivars) ->
-    trans_let_function(Fbs, Body, Name, Ar, New, Ivars);
-trans_expr(['letrec-function',Fbs|Body], Name, Ar, New, Ivars) ->
-    trans_letrec_function(Fbs, Body, Name, Ar, New, Ivars);
-%% Core control special forms.
-trans_expr([progn|Body], Name, Ar, New, Ivars) ->
-    [progn|trans_exprs(Body, Name, Ar, New, Ivars)];
-trans_expr(['if'|Body], Name, Ar, New, Ivars) ->
-    ['if'|trans_exprs(Body, Name, Ar, New, Ivars)];
-trans_expr(['case',Expr|Cls], Name, Ar, New, Ivars) ->
-    trans_case(Expr, Cls, Name, Ar, New, Ivars);
-trans_expr(['receive'|Cls], Name, Ar, New, Ivars) ->
-    ['receive'|trans_cls(Cls, Name, Ar, New, Ivars)];
-trans_expr(['catch'|Body], Name, Ar, New, Ivars) ->
-    ['catch'|trans_exprs(Body, Name, Ar, New, Ivars)];
-trans_expr(['try'|Body],  Name, Ar, New, Ivars) ->
-    trans_try(Body, Name, Ar, New, Ivars);
-trans_expr([funcall|Body], Name, Ar, New, Ivars) ->
-    [funcall|trans_exprs(Body, Name, Ar, New, Ivars)];
-%% List/binary comprehensions.
-trans_expr(['lc',Qs,E], Name, Ar, New, Ivars) ->
-    trans_comp('lc', Qs, E, Name, Ar, New, Ivars);
-trans_expr(['list-comp',Qs,E], Name, Ar, New, Ivars) ->
-    trans_comp('list-comp', Qs, E, Name, Ar, New, Ivars);
-trans_expr(['bc',Qs,E], Name, Ar, New, Ivars) ->
-    trans_comp('bc', Qs, E, Name, Ar, New, Ivars);
-trans_expr(['binary-comp',Qs,E], Name, Ar, New, Ivars) ->
-    trans_comp('binary-comp', Qs, E, Name, Ar, New, Ivars);
-%% General cases.
-trans_expr([call|Body], Name, Ar, New, Ivars) ->
-    [call|trans_exprs(Body, Name, Ar, New, Ivars)];
-trans_expr([Fun|Args0], Name, Ar, New, Ivars) when is_atom(Fun) ->
-    %% Most of the core data special forms can be handled here as well.
-    Far = length(Args0),
-    Args1 = trans_exprs(Args0, Name, Ar, New, Ivars),
-    if Fun =:= Name,
-       Far =:= Ar -> [New|Args1 ++ Ivars];
-       true -> [Fun|Args1]
-    end;
-trans_expr(Lit, _, _, _, _) -> Lit.
-
-trans_exprs(Exprs, Name, Ar, New, Ivars) ->
-    Fun = fun (E) -> trans_expr(E, Name, Ar, New, Ivars) end,
-    lists:map(Fun, Exprs).
-
-trans_bitsegs(Segs, Name, Ar, New, Ivars) ->
-    Fun = fun (Seg) -> trans_bitseg(Seg, Name, Ar, New, Ivars) end,
-    lists:map(Fun, Segs).
-
-trans_bitseg([Val0|Specs0], Name, Ar, New, Ivars) ->
-    Val1 = trans_expr(Val0, Name, Ar, New, Ivars),
-    Fun = fun ([size,E]) -> [size,trans_expr(E, Name, Ar, New, Ivars)] end,
-    Specs1 = lists:map(Fun, Specs0),
-    [Val1|Specs1];
-trans_bitseg(Seg, Name, Ar, New, Ivars) ->
-    trans_expr(Seg, Name, Ar, New, Ivars) .
-
-%% trans_rec_fields(Fields, Name, Arity, NewName, ImportedVars) -> Fields.
-%% trans_rec_args(Args, Name, Arity, NewName, ImportedVars) -> Args.
-
-%% trans_rec_fields([[F,V|Type]|Fs], Name, Ar, New, Ivars) ->
-%%     Tv = trans_expr(V, Name, Ar, New, Ivars),
-%%     Tfs = trans_rec_fields(Fs,  Name, Ar, New, Ivars),
-%%     [[F,Tv|Type]|Tfs];
-%% trans_rec_fields([F|Fs], Name, Ar, New, Ivars) ->
-%%     Tfs = trans_rec_fields(Fs,  Name, Ar, New, Ivars),
-%%     [F|Tfs];
-%% trans_rec_fields([], _, _, _, _) -> [].
-
-trans_rec_args([F,V|As], Name, Ar, New, Ivars) ->
-    Tv = trans_expr(V, Name, Ar, New, Ivars),
-    Tas = trans_rec_args(As, Name, Ar, New, Ivars),
-    [F,Tv|Tas];
-trans_rec_args([], _, _, _, _) -> [].
-
-trans_cls(Cls, Name, Ar, New, Ivars) ->
-    Fun = fun (Cl) -> trans_cl(Cl, Name, Ar, New, Ivars) end,
-    lists:map(Fun, Cls).
-
-%% trans_cl(Clause, Name, Arity, NewName, ImportedVars) -> Clause.
-%%  We know that there are no interesting functions in the guard.
-
-trans_cl([Pat,['when'|_]=G|Body], Name, Ar, New, Ivars) ->
-    [Pat,G|trans_exprs(Body, Name, Ar, New, Ivars)];
-trans_cl([Pat|Body], Name, Ar, New, Ivars) ->
-    [Pat|trans_exprs(Body, Name, Ar, New, Ivars)].
-
-trans_let(Vbs0, Body0, Name, Ar, New, Ivars) ->
-    Fun = fun ([Pat,['when'|_]=G,Expr0]) ->
-                  Expr1 = trans_expr(Expr0, Name, Ar, New, Ivars),
-                  [Pat,G,Expr1];
-              ([Pat,Expr0]) ->
-                  Expr1 = trans_expr(Expr0, Name, Ar, New, Ivars),
-                  [Pat,Expr1]
-          end,
-    Vbs1 = lists:map(Fun, Vbs0),
-    Body1 = trans_exprs(Body0, Name, Ar, New, Ivars),
-    ['let',Vbs1|Body1].
-
-trans_let_function(Fbs0, Body0, Name, Ar, New, Ivars) ->
-    Fbs1 = trans_let_fbs(Fbs0, Name, Ar, New, Ivars),
-    Body1 = trans_exprs(Body0, Name, Ar, New, Ivars),
-    ['let-function',Fbs1|Body1].
-
-trans_letrec_function(Fbs0, Body0, Name, Ar, New, Ivars) ->
-    Fbs1 = trans_let_fbs(Fbs0, Name, Ar, New, Ivars),
-    Body1 = trans_exprs(Body0, Name, Ar, New, Ivars),
-    ['letrec-function',Fbs1|Body1].
-
-trans_let_fbs(Fbs, Name, Ar, New, Ivars) ->
-    Fun = fun ([F,Def]) -> [F,trans_expr(Def, Name, Ar, New, Ivars)] end,
-    lists:map(Fun, Fbs).
-
-trans_case(Expr0, Cls0, Name, Ar, New, Ivars) ->
-    Expr1 = trans_expr(Expr0, Name, Ar, New, Ivars),
-    Cls1 = trans_cls(Cls0, Name, Ar, New, Ivars),
-    ['case',Expr1|Cls1].
-
-%% trans_try(TryBody, Name, Arity, NewName, ImportedVars) -> Try.
-%%  Step down the try body doing each section separately.
-
-trans_try(Try0, Name, Ar, New, Ivars) ->
-    Fun = fun (T) -> trans_try_1(T, Name, Ar, New, Ivars) end,
-    Try1 = lists:map(Fun, Try0),
-    ['try'|Try1].
-
-trans_try_1(['case'|Case0], Name, Ar, New, Ivars) ->
-    Case1 = trans_cls(Case0, Name, Ar, New, Ivars),
-    ['case'|Case1];
-trans_try_1(['catch'|Catch0], Name, Ar, New, Ivars) ->
-    Catch1 = trans_cls(Catch0, Name, Ar, New, Ivars),
-    ['catch'|Catch1];
-trans_try_1(['after'|After0], Name, Ar, New, Ivars) ->
-    After1 = trans_exprs(After0, Name, Ar, New, Ivars),
-    ['after'|After1];
-trans_try_1(E, Name, Ar, New, Ivars) ->         %The try expression.
-    trans_expr(E, Name, Ar, New, Ivars).
-                       
-func_arity([lambda,Args|_]) -> length(Args);
-func_arity(['match-lambda',[Pats|_]|_]) ->
-    length(Pats).
-
-%% trans_comp(Comprehension, Qualifiers, Expr,
-%%            OldName, Arity, NewName, ImportedVars) ->
-%%      Expr.
-%%  Translate a list/binary comprehenesion.
-
-trans_comp(Comp, Qs0, E0, Name, Ar, New, Ivars) ->
-    E1 = trans_expr(E0, Name, Ar, New, Ivars),
-    Qs1 = trans_comp_quals(Qs0, Name, Ar, New, Ivars),
-    [Comp,Qs1,E1].
-
-trans_comp_quals(Qs, Name, Ar, New, Ivars) ->
-    lists:map(fun (Q) ->
-		      trans_comp_qual(Q, Name, Ar, New, Ivars)
-	      end, Qs).
-
-trans_comp_qual(['<-',Pat,E0], Name, Ar, New, Ivars) ->
-    E1 = trans_expr(E0, Name, Ar, New, Ivars),
-    io:format("tcq ~p ~p\n", [E0,E1]),
-    ['<-',Pat,E1];
-trans_comp_qual(['<-',Pat,Guard,E0], Name, Ar, New, Ivars) ->
-    E1 = trans_expr(E0, Name, Ar, New, Ivars),
-    ['<-',Pat,Guard,E1];
-trans_comp_qual(['<=',Pat,E0], Name, Ar, New, Ivars) ->
-    E1 = trans_expr(E0, Name, Ar, New, Ivars),
-    ['<=',Pat,E1];
-trans_comp_qual(['<=',Pat,Guard,E0], Name, Ar, New, Ivars) ->
-    E1 = trans_expr(E0, Name, Ar, New, Ivars),
-    ['<=',Pat,Guard,E1];
-trans_comp_qual(Test, Name, Ar, New, Ivars) ->
-    trans_expr(Test, Name, Ar, New, Ivars).
+lift_comp_qual(Test, Lifts, Lds, St) ->
+    lift_expr(Test, Lifts, Lds, St).
 
 %% new_local_fun_name(Name, Arity, State) -> {FunName,State}.
 %%  Create a name for a local function. The name has a similar basic
@@ -692,7 +566,8 @@ ivars_expr(Var, Kvars, Ivars) when is_atom(Var) ->
         true -> Ivars;
         false -> ordsets:add_element(Var, Ivars)
     end;
-ivars_expr(_Lit, _Kvars, Ivars) -> Ivars.       %All literals
+%% Everything else is a literal.
+ivars_expr(_Lit, _Kvars, Ivars) -> Ivars.
 
 ivars_exprs(Exprs, Kvars, Ivars) ->
     Fun = fun (E, Ivs) -> ivars_expr(E, Kvars, Ivs) end,
@@ -807,9 +682,9 @@ ivars_comp(Qs, E, Kvars0, Ivars0) ->
 
 ivars_comp_quals(Qs, Kvars, Ivars) ->
     lists:foldl(fun (Q, {Kvars0,Ivars0}) ->
-			{Kvars1,Ivars1} = ivars_comp_qual(Q, Kvars0, Ivars0),
-			{Kvars1,Ivars1}
-		end, {Kvars,Ivars}, Qs).
+                        {Kvars1,Ivars1} = ivars_comp_qual(Q, Kvars0, Ivars0),
+                        {Kvars1,Ivars1}
+                end, {Kvars,Ivars}, Qs).
 
 ivars_comp_qual(['<-',Pat,Gen], Kvars, Ivars) ->
     ivars_comp_qual(Pat, [], Gen, Kvars, Ivars);
@@ -849,11 +724,11 @@ ivars_pats(Pats, Pvars) ->
 
 %% mapfoldl2(Fun, Acc1, Acc2, List) -> {List,Acc1,Acc2}.
 
-mapfoldl2(Fun, A0, B0, [E0|Es0]) ->
-    {E1,A1,B1} = Fun(E0, A0, B0),
-    {Es1,A2,B2} = mapfoldl2(Fun, A1, B1, Es0),
-    {[E1|Es1],A2,B2};
-mapfoldl2(_, A, B, []) -> {[],A,B}.
+%% mapfoldl2(Fun, A0, B0, [E0|Es0]) ->
+%%     {E1,A1,B1} = Fun(E0, A0, B0),
+%%     {Es1,A2,B2} = mapfoldl2(Fun, A1, B1, Es0),
+%%     {[E1|Es1],A2,B2};
+%% mapfoldl2(_, A, B, []) -> {[],A,B}.
 
 %% test(Which) -> Sexpr.
 
