@@ -1,4 +1,5 @@
-%% Copyright (c) 2008-2022 Robert Virding
+%% -*- mode: erlang; indent-tabs-mode: nil -*-
+%% Copyright (c) 2008-2024 Robert Virding
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -31,13 +32,13 @@
          run_strings/1,run_strings/2,run_string/1,run_string/2,
          new_state/2,new_state/3,upd_state/3]).
 
-%% The shell commands which generally callable.
--export([c/1,c/2,cd/1,ec/1,ec/2,ep/1,ep/2,epp/1,epp/2,help/0,h/1,h/2,h/3,
-         i/0,i/1,i/3,l/1,ls/1,clear/0,m/0,m/1,pid/3,p/1,p/2,pp/1,pp/2,pwd/0,
-         q/0,flush/0,regs/0,exit/0]).
+%% Useful for the LFE rebar3 plugin
+-export([banner/0,banner/1,banner/2]).
 
--import(orddict, [store/3,find/2]).
--import(lists, [reverse/1,foreach/2]).
+%% The shell commands which generally callable.
+-export([c/1,c/2,cd/1,ec/1,ec/2,ep/1,ep/2,epp/1,epp/2,flush/0,help/0,
+         h/1,h/2,h/3,i/0,i/1,i/3,l/1,ls/1,clear/0,m/0,m/1,memory/0,memory/1,
+         nregs/0,pid/3,p/1,p/2,pp/1,pp/2,pwd/0,q/0,regs/0,uptime/0,exit/0]).
 
 -include("lfe.hrl").
 -include("lfe_docs.hrl").
@@ -97,7 +98,7 @@ run_string(String, St0) ->
     St1 = upd_state(String, [], St0),
     case read_script_string(String) of
         {ok,Forms} ->
-            run_loop([Forms], St1);
+            run_loop(Forms, St1);
         {error,E} ->
             slurp_errors("lfe", [E]),
             {error,St1}
@@ -152,15 +153,15 @@ shell_eval(Form, Eval0, St0) ->
         {eval_value,Eval0,_Value,St1} ->
             %% We actually don't use the returned value here.
             {Eval0,St1};
-        {eval_error,Eval0,Class} ->
+        {eval_error,Eval0,{Class,Reason,StackTrace}} ->
             %% Eval errored out, get the exit signal.
-            receive {'EXIT',Eval0,{Reason,Stk}} -> ok end,
-            report_exception(Class, Reason, Stk),
+            receive {'EXIT',Eval0,normal} -> ok end,
+            report_exception(Class, Reason, StackTrace),
             Eval1 = start_eval(St0),
             {Eval1,St0};
-        {'EXIT',Eval0,Error} ->
+        {'EXIT',Eval0,Reason} ->
             %% Eval exited or was killed
-            report_exception(error, Error, []),
+            report_exception(exit, Reason, []),
             Eval1 = start_eval(St0),
             {Eval1,St0}
     end.
@@ -198,18 +199,15 @@ user_prompt () ->
         _ -> ["lfe> "]
     end.
 
-report_exception(Class, Reason, Stk) ->
+report_exception(Class, Reason, Stack) ->
     %% Use LFE's simplified version of erlang shell's error
     %% reporting but which LFE prettyprints data.
-    Sf = fun ({M,_F,_A}) ->                     %Pre R15
-                 %% Don't want to see these in stacktrace.
-                 (M == lfe_eval) or (M == ?MODULE);
-             ({M,_F,_A,_L}) ->                  %R15 and later
+    Skip = fun (M, _F, _A) ->
                  %% Don't want to see these in stacktrace.
                  (M == lfe_eval) or (M == ?MODULE)
-         end,
-    Ff = fun (T, I) -> lfe_io:prettyprint1(T, 15, I, 80) end,
-    Cs = lfe_lib:format_exception(Class, Reason, Stk, Sf, Ff, 1),
+           end,
+    Format = fun (T, I) -> lfe_io:prettyprint1(T, 15, I, 80) end,
+    Cs = lfe_error:format_exception(Class, Reason, Stack, Skip, Format, 1),
     io:put_chars("** " ++ Cs),                  %Make it more note worthy
     io:nl().
 
@@ -242,17 +240,23 @@ read_expression_1(Rdr, Eval, St) ->
             read_expression_1(Rdr, start_eval(St), St)
     end.
 
-make_banner() ->
-    [io_lib:format(
-       ?GRN("   ..-~~") ++ ?YLW(".~~_") ++ ?GRN("~~---..") ++ "\n" ++
+banner() ->
+    banner(lfe:version()).
+
+banner(Vsn) ->
+    banner(Vsn, quit_message()).
+
+banner(Vsn, QuitMsg) ->
+    ?GRN("   ..-~") ++ ?YLW(".~_") ++ ?GRN("~---..") ++ "\n" ++
        ?GRN("  (      ") ++ ?YLW("\\\\") ++ ?GRN("     )") ++ "    |   A Lisp-2+ on the Erlang VM\n" ++
        ?GRN("  |`-.._") ++ ?YLW("/") ++ ?GRN("_") ++ ?YLW("\\\\") ++ ?GRN("_.-':") ++ "    |   Type " ++ ?GRN("(help)") ++ " for usage info.\n" ++
        ?GRN("  |         ") ++ ?RED("g") ++ ?GRN(" |_ \\") ++  "   |\n" ++
        ?GRN("  |        ") ++ ?RED("n") ++ ?GRN("    | |") ++   "  |   Docs: " ++ ?BLU("http://docs.lfe.io/") ++ "\n" ++
        ?GRN("  |       ") ++ ?RED("a") ++ ?GRN("    / /") ++   "   |   Source: " ++ ?BLU("http://github.com/lfe/lfe") ++ "\n" ++
        ?GRN("   \\     ") ++ ?RED("l") ++ ?GRN("    |_/") ++  "    |\n" ++
-       ?GRN("    \\   ") ++ ?RED("r") ++ ?GRN("     /") ++  "      |   LFE v~s ~s\n" ++
-       ?GRN("     `-") ++ ?RED("E") ++ ?GRN("___.-'") ++ "\n\n", [get_lfe_version(), get_abort_message()])].
+       ?GRN("    \\   ") ++ ?RED("r") ++ ?GRN("     /") ++  "      |   LFE v" ++
+        Vsn ++ " " ++  QuitMsg ++ "\n" ++
+       ?GRN("     `-") ++ ?RED("E") ++ ?GRN("___.-'") ++ "\n\n".
 
 display_banner() ->
     %% When LFE is called with -noshell, we want to skip the banner. Also, there may be
@@ -260,22 +264,18 @@ display_banner() ->
     %% thus we want to support both use cases.
     case init:get_argument(noshell) of
         error -> case init:get_argument(nobanner) of
-                    error -> io:put_chars(make_banner());
+                    error -> io:put_chars(banner());
                     _ -> false
                  end;
         _ -> false
     end.
 
-get_abort_message() ->
+quit_message() ->
     %% We can update this later to check for env variable settings for
     %% shells that require a different control character to abort, such
     %% as jlfe.
     "(abort with ^G)".
 
-get_lfe_version() ->
-    {ok, [App]} = file:consult(code:where_is_file("lfe.app")),
-    proplists:get_value(vsn, element(3, App)).
-
 %% new_state(ScriptName, Args [,Env]) -> State.
 %%  Generate a new shell state with all the default functions, macros
 %%  and variables.
@@ -330,6 +330,7 @@ add_shell_functions(Env0) ->
           {h,1,[lambda,[m], [':',lfe_shell,h,m]]},
           {h,2,[lambda,[m,f], [':',lfe_shell,h,m,f]]},
           {h,3,[lambda,[m,f,a], [':',lfe_shell,h,m,f,a]]},
+
           {help,0,[lambda,[],[':',lfe_shell,help]]},
           {i,0,[lambda,[],[':',lfe_shell,i]]},
           {i,1,[lambda,[ps],[':',lfe_shell,i,ps]]},
@@ -344,6 +345,10 @@ add_shell_functions(Env0) ->
           {q,0,[lambda,[],[':',lfe_shell,exit]]},
           {flush,0,[lambda,[],[':',lfe_shell,flush]]},
           {regs,0,[lambda,[],[':',lfe_shell,regs]]},
+          {nregs,0,[lambda,[],[':',lfe_shell,nregs]]},
+          {memory,0,[lambda,[],[':',lfe_shell,memory]]},
+          {memory,1,[lambda,[t],[':',lfe_shell,memory,t]]},
+          {uptime,0,[lambda,[],[':',lfe_shell,uptime]]},
           {exit,0,[lambda,[],[':',lfe_shell,exit]]}
          ],
     %% Any errors here will crash shell startup!
@@ -423,16 +428,19 @@ eval_form(Form, Shell, St0) ->
         St2
     catch
         exit:normal -> exit(normal);
-        ?CATCH(Class, Reason, Stack)
+        ?CATCH(Class, Reason, StackTrace)
             %% We don't want the ERROR REPORT generated by the
             %% emulator. Note: exit(kill) needs nothing special.
-            Shell ! {eval_error,self(),Class},
-            E = nocatch(Class, {Reason,Stack}),
-            exit(E)
+            {links,LPs} = process_info(self(), links),
+            ER = nocatch(Class, {Reason,StackTrace}),
+            lists:foreach(fun(P) -> exit(P, ER) end, LPs -- [Shell]),
+            Shell ! {eval_error,self(),{Class,Reason,StackTrace}},
+            %% Note we are linked to Shell so we don't crash here.
+            exit(normal)
     end.
 
-nocatch(throw, {Term,Stack}) ->
-    {{nocatch,Term},Stack};
+%% nocatch(throw, {Term,Stack}) ->
+%%     {{nocatch,Term},Stack};
 nocatch(_, Reason) -> Reason.
 
 %% eval_form(Form, State) -> {Value,State}.
@@ -502,10 +510,10 @@ list_errors(Es) -> list_ews("~w: ~s\n", Es).
 list_warnings(Ws) -> list_ews("~w: Warning: ~s\n", Ws).
 
 list_ews(Format, Ews) ->
-    foreach(fun ({L,M,E}) ->
-                    Cs = M:format_error(E),
-                    lfe_io:format(Format, [L,Cs])
-            end, Ews).
+    lists:foreach(fun ({L,M,E}) ->
+                          Cs = M:format_error(E),
+                          lfe_io:format(Format, [L,Cs])
+                  end, Ews).
 
 %% set(Args, State) -> {Result,State}.
 
@@ -537,7 +545,7 @@ set_1(Pat, Guard, Exp, #state{curr=Ce0}=St) ->
 %%  to get hold of the macros. We call the compiler directly but don't
 %%  make the LFE-EXPAND-EXPORTED-MACRO/3 function.
 
--record(slurp, {mod,funs=[],imps=[]}).          %For slurping
+-record(slurp, {mod,funs=[],imps=[],recs=[]}).  %For slurping
 
 unslurp(St0) ->
     St1 = case St0#state.slurp of
@@ -563,7 +571,7 @@ slurp_1(Name, Ce) ->
         {ok,Mod,Fs,Env0,Ws} ->
             slurp_warnings(Ws),
             %% Collect functions and imports.
-            Sl0 = #slurp{mod=Mod,funs=[],imps=[]},
+            Sl0 = #slurp{mod=Mod,funs=[],imps=[],recs=[]},
             Sl1 = lists:foldl(fun collect_module/2, Sl0, Fs),
             %% Add imports to environment.
             Ifun = fun ({M,Is}, Env) ->
@@ -576,15 +584,19 @@ slurp_1(Name, Ce) ->
             Env2 = foldl(fun ({N,Ar,Def}, Env) ->
                                  lfe_eval:add_dynamic_func(N, Ar, Def, Env)
                          end, Env1, Sl1#slurp.funs),
-            {ok,Mod,lfe_env:add_env(Env2, Ce)};
+            %% Add records to the environment.
+            Env3 = foldl(fun ({R,Rfs}, Env) ->
+                                 lfe_env:add_record(R, Rfs, Env)
+                         end, Env2, Sl1#slurp.recs),
+            {ok,Mod,lfe_env:add_env(Env3, Ce)};
         {error,Mews,Es,Ws} ->
             slurp_errors(Es),
             slurp_warnings(Ws),
             %% Now the errors and warnings for each module.
-            foreach(fun ({error,Mes,Mws}) ->
-                            slurp_errors(Mes),
-                            slurp_warnings(Mws)
-                    end, Mews),
+            lists:foreach(fun ({error,Mes,Mws}) ->
+                                  slurp_errors(Mes),
+                                  slurp_warnings(Mws)
+                          end, Mews),
             error
     end.
 
@@ -623,8 +635,10 @@ collect_module({['extend-module',Meta,Atts],_}, Sl0) ->
 collect_module({['define-function',F,_Meta,Def],_}, #slurp{funs=Fs}=Sl) ->
     Ar = function_arity(Def),
     Sl#slurp{funs=[{F,Ar,Def}|Fs]};
+collect_module({['define-record',R,Fs],_}, #slurp{recs=Rs}=Sl) ->
+    Sl#slurp{recs=[{R,Fs}|Rs]};
+%% Ignore other forms, type and spec defs.
 collect_module({_,_}, Sl) ->
-    %% Ignore other forms, type and spec defs.
     Sl.
 
 collect_meta(_, St) -> St.
@@ -639,37 +653,37 @@ collect_imps(Is, St) ->
     foldl(fun (I, S) -> collect_imp(I, S) end, St, Is).
 
 collect_imp(['from',Mod|Fs], St) ->
-    collect_imp(fun ([F,A], Imps) -> store({F,A}, F, Imps) end,
+    collect_imp(fun ([F,A], Imps) -> orddict:store({F,A}, F, Imps) end,
                 Mod, St, Fs);
 collect_imp(['rename',Mod|Rs], St) ->
-    collect_imp(fun ([[F,A],R], Imps) -> store({F,A}, R, Imps) end,
+    collect_imp(fun ([[F,A],R], Imps) -> orddict:store({F,A}, R, Imps) end,
                 Mod, St, Rs);
 collect_imp(_, St) -> St.                       %Ignore everything else
 
 collect_imp(Fun, Mod, St, Fs) ->
     Imps0 = safe_fetch(Mod, St#slurp.imps, []),
     Imps1 = foldl(Fun, Imps0, Fs),
-    St#slurp{imps=store(Mod, Imps1, St#slurp.imps)}.
+    St#slurp{imps=orddict:store(Mod, Imps1, St#slurp.imps)}.
 
 %% slurp_errors([File, ]Errors) -> ok.
 %% slurp_warnings([File, ]Warnings) -> ok.
 %%  Print errors and warnings.
 
 slurp_errors(Errors) ->
-    foreach(fun ({File,Es}) -> slurp_errors(File, Es) end, Errors).
+    lists:foreach(fun ({File,Es}) -> slurp_errors(File, Es) end, Errors).
 
 slurp_errors(File, Es) -> slurp_ews(File, "~s:~w: ~s\n", Es).
 
 slurp_warnings(Warnings) ->
-    foreach(fun ({File,Ws}) -> slurp_warnings(File, Ws) end, Warnings).
+    lists:foreach(fun ({File,Ws}) -> slurp_warnings(File, Ws) end, Warnings).
 
 slurp_warnings(File, Es) -> slurp_ews(File, "~s:~w: Warning: ~s\n", Es).
 
 slurp_ews(File, Format, Ews) ->
-    foreach(fun ({Line,Mod,Error}) ->
-                    Cs = Mod:format_error(Error),
-                    lfe_io:format(Format, [File,Line,Cs])
-            end, Ews).
+    lists:foreach(fun ({Line,Mod,Error}) ->
+                          Cs = Mod:format_error(Error),
+                          lfe_io:format(Format, [File,Line,Cs])
+                  end, Ews).
 
 %% run_file(Args, State) -> {Value,State}.
 %%  Run the shell expressions in a file. Abort on errors and only
@@ -713,7 +727,9 @@ read_script_string(String) ->
 %% run_loop(Forms, State) -> {Value,State}.
 %% run_loop(Forms PrevValue, State) -> {Value,State}.
 
-run_loop(Fs, St) -> run_loop(Fs, [], St).
+run_loop(Fs, St) ->
+    run_loop(Fs, [], St).
+
 
 run_loop([F|Fs], _, St0) ->
     Ce1 = lfe_env:add_vbinding('-', F, St0#state.curr),
@@ -725,7 +741,7 @@ run_loop([], Value, St) -> {Value,St}.
 %% safe_fetch(Key, Dict, Default) -> Value.
 
 safe_fetch(Key, D, Def) ->
-    case find(Key, D) of
+    case orddict:find(Key, D) of
         {ok,Val} -> Val;
         error -> Def
     end.
@@ -827,12 +843,16 @@ help() ->
                    "(ls dir)       -- list files in directory <dir>\n"
                    "(m)            -- which modules are loaded\n"
                    "(m mod)        -- information about module <mod>\n"
+                   "(memory)       -- memory allocation information\n"
+                   "(memory t)     -- memory allocation information of type <t>\n"
                    "(p expr)       -- print a term\n"
                    "(pp expr)      -- pretty print a term\n"
                    "(pid x y z)    -- convert x, y, z to a pid\n"
                    "(pwd)          -- print working directory\n"
                    "(q)            -- quit - shorthand for init:stop/0\n"
                    "(regs)         -- information about registered processes\n"
+                   "(nregs)        -- information about all registered processes\n"
+                   "(uptime)       -- print node uptime\n"
                    "\n"
                    "LFE shell built-in forms\n\n"
                    "(reset-environment)             -- reset the environment to its initial state\n"
@@ -863,7 +883,7 @@ i(X, Y, Z) -> c:i(X, Y, Z).
 %%  Load the modules.
 
 l(Ms) ->
-    foreach(fun (M) -> c:l(M) end, Ms).
+    lists:map(fun (M) -> c:l(M) end, Ms).
 
 %% ls(Dir) -> ok.
 
@@ -892,7 +912,7 @@ mformat(S1, S2) ->
     lfe_io:format(Fstr, [S1,S2]).
 
 m(Ms) ->
-    foreach(fun (M) -> print_module(M) end, Ms).
+    lists:foreach(fun (M) -> print_module(M) end, Ms).
 
 print_module(M) ->
     Info = M:module_info(),
@@ -1018,10 +1038,26 @@ flush() -> c:flush().
 
 regs() -> c:regs().
 
+%% nregs() -> ok.
+
+nregs() -> c:nregs().
+
 %% exit() -> ok.
 
 exit() -> c:q().
 
+%% memory() -> ok.
+
+memory() -> c:memory().
+
+%% memory(Type) -> ok.
+
+memory(Type) -> c:memory(Type).
+
+%% uptime() -> ok.
+
+uptime() -> c:uptime().
+
 %% doc(Mod) -> ok | {error,Error}.
 %% doc(Mod, Func) -> ok | {error,Error}.
 %% doc(Mod, Func, Arity) -> ok | {error,Error}.
@@ -1131,10 +1167,22 @@ paged_output(Pitem, Curr, Limit, [Item|Items]) ->
 paged_output(_, _, _, []) -> ok.
 
 more() ->
-    case io:get_line('More (y/n)? ') of
+    case get_line('More (y/n)? (y)', "y\n") of
         "y\n" -> more;
         "c\n" -> more;
         "n\n" -> less;
         "q\n" -> less;
         _ -> more()
     end.
+
+get_line(P, Default) ->
+    case line_string(io:get_line(P)) of
+        "\n" -> Default;
+        L -> L
+    end.
+
+%% If the standard input is set to binary mode
+%% convert it to a list so we can properly match.
+line_string(Binary) when is_binary(Binary) ->
+    unicode:characters_to_list(Binary);
+line_string(Other) -> Other.
